from __future__ import annotations

import typing as t

import attrs

from unkey import constants as c

__all__ = ("CompiledRoute", "Route")


@attrs.define(weakref_slot=False)
class CompiledRoute:
    """A route that has been compiled to include uri variables.

    Args:
        route: The route itself.
        uri: The endpoint for this route.
    """

    route: Route
    """The route itself."""

    uri: str
    """The routes uri endpoint."""

    params: t.Dict[str, t.Union[str, int]] = attrs.field(init=False)
    """The query params for the route."""

    def __init__(self, route: Route, uri: str) -> None:
        self.route = route
        self.uri = uri
        self.params = {}

    @property
    def method(self) -> str:
        """The routes method, i.e. GET, POST..."""
        return self.route.method

    def with_params(self, params: t.Dict[str, t.Any]) -> CompiledRoute:
        """Adds additional query params to this compiled route.

        Args:
            params: The query params to compile.

        Returns:
            The compiled route for chained calls.
        """
        if params:
            self.params.update(params)

        return self


@attrs.define(weakref_slot=False)
class Route:
    """A route that has not been compiled yet."""

    method: str
    """The request method to use."""

    uri: str
    """The request uri."""

    def compile(self, *args: t.Union[str, int]) -> CompiledRoute:
        """Turn this route into a compiled route.

        Args:
            *args: The arguments to insert into the uri.

        Returns:
            The compiled route.
        """
        compiled = CompiledRoute(self, self.uri)

        for arg in args:
            compiled.uri = compiled.uri.replace(r"{}", str(arg), 1)

        return compiled


# Keys
CREATE_KEY: t.Final[Route] = Route(c.POST, "/keys.createKey")
VERIFY_KEY: t.Final[Route] = Route(c.POST, "/keys.verifyKey")
REVOKE_KEY: t.Final[Route] = Route(c.POST, "/keys.deleteKey")
UPDATE_KEY: t.Final[Route] = Route(c.POST, "/keys.updateKey")
UPDATE_REMAINING: t.Final[Route] = Route(c.POST, "/keys.updateRemaining")
GET_KEY: t.Final[Route] = Route(c.GET, "/keys.getKey")

# Apis
GET_API: t.Final[Route] = Route(c.GET, "/apis.getApi")
GET_KEYS: t.Final[Route] = Route(c.GET, "/apis.listKeys")



from __future__ import annotations

import platform
from pathlib import Path

from unkey import __git_sha__
from unkey import __version__


def _main() -> None:
    """Prints package/system info and exits."""
    path = Path(__file__).parent.absolute()
    py_impl = platform.python_implementation()
    py_ver = platform.python_version()
    py_c = platform.python_compiler()
    p = platform.uname()

    print(f"unkey.py v{__version__} from {__git_sha__}")
    print(f"@ {path}")
    print(f"{py_impl} {py_ver} {py_c}")
    print(f"{p.system} {p.node} {p.release} {p.machine}")
    print(p.version)


if __name__ == "__main__":
    _main()



from __future__ import annotations

from typing import Final

import unkey

__all__ = ()

API_BASE_URL: Final[str] = "https://api.unkey.dev"
USER_AGENT: Final[str] = f"unkey.py@v{unkey.__version__}"

GET: Final[str] = "GET"
PUT: Final[str] = "PUT"
POST: Final[str] = "POST"
PATCH: Final[str] = "PATCH"
DELETE: Final[str] = "DELETE"



from __future__ import annotations

from typing import Final

__packagename__: Final[str] = "unkey.py"
__version__: Final[str] = "0.7.1"
__author__: Final[str] = "Jonxslays"
__copyright__: Final[str] = "2023-present Jonxslays"
__description__: Final[str] = "An asynchronous Python SDK for unkey.dev."
__url__: Final[str] = "https://github.com/Jonxslays/unkey.py"
__docs__: Final[str] = "https://jonxslays.github.io/unkey.py"
__repository__: Final[str] = __url__
__license__: Final[str] = "GPL-3.0"
__git_sha__: Final[str] = "[024513f]"

from . import client
from . import decorators
from . import constants
from . import errors
from . import models
from . import result
from . import routes
from . import serializer
from . import services
from . import undefined
from .client import *
from .decorators import *
from .errors import *
from .models import *
from .result import *
from .routes import *
from .serializer import *
from .services import *
from .undefined import *

__all__ = (
    "client",
    "constants",
    "decorators",
    "errors",
    "models",
    "protected",
    "result",
    "routes",
    "serializer",
    "services",
    "undefined",
    "Api",
    "ApiKey",
    "ApiKeyList",
    "ApiKeyMeta",
    "ApiKeyVerification",
    "ApiService",
    "BaseEnum",
    "BaseError",
    "BaseModel",
    "BaseService",
    "Client",
    "CompiledRoute",
    "Err",
    "ErrorCode",
    "HttpResponse",
    "HttpService",
    "KeyService",
    "MissingRequiredArgument",
    "Ok",
    "Ratelimit",
    "RatelimitState",
    "RatelimitType",
    "Refill",
    "RefillInterval",
    "Result",
    "Route",
    "Serializer",
    "UndefinedNoneOr",
    "UndefinedOr",
    "UnwrapError",
    "UNDEFINED",
    "UpdateOp",
)



from __future__ import annotations

import typing as t

import attrs

from .base import BaseModel
from .keys import ApiKeyMeta

__all__ = ("Api", "ApiKeyList")


@attrs.define(init=False, weakref_slot=False)
class Api(BaseModel):
    """Data representing a particular ratelimit."""

    id: str
    """The id for this api."""

    name: str
    """The name of the api."""

    workspace_id: str
    """The id for the workspace this api belongs to."""


@attrs.define(init=False, weakref_slot=False)
class ApiKeyList(BaseModel):
    """Data representing keys for an api."""

    cursor: t.Optional[str]
    """The cursor indicating the last key that was returned."""

    keys: t.List[ApiKeyMeta]
    """A list of keys associated with the api."""

    total: int
    """The total number of keys associated with the api."""



from __future__ import annotations

import abc
import typing as t
from datetime import datetime
from datetime import timedelta

from unkey import undefined

if t.TYPE_CHECKING:  # pragma: nocover
    from unkey import serializer

    from . import HttpService

__all__ = ("BaseService",)


class BaseService(abc.ABC):
    """The base service all API services inherit from.

    Args:
        http_service: The http service to use for requests.

        serializer: The serializer to use for handling incoming
            JSON data from the API.
    """

    __slots__ = ("_http", "_serializer")

    def __init__(self, http_service: HttpService, serializer: serializer.Serializer) -> None:
        self._http = http_service
        self._serializer = serializer

    def _generate_map(self, **kwargs: t.Any) -> t.Dict[str, t.Any]:
        return {k: v for k, v in kwargs.items() if v is not undefined.UNDEFINED}

    def _expires_in(
        self, *, milliseconds: int = 0, seconds: int = 0, minutes: int = 0, days: int = 0
    ) -> undefined.UndefinedOr[int]:
        if not any({milliseconds, seconds, minutes, days}):
            return undefined.UNDEFINED

        delta = timedelta(days=days, minutes=minutes, seconds=seconds, milliseconds=milliseconds)
        return int((datetime.now() + delta).timestamp()) * 1000



from __future__ import annotations

import typing as t
from datetime import datetime

from unkey import models

__all__ = ("Serializer",)

T = t.TypeVar("T")
DictT = t.Dict[str, t.Any]
TransformT = t.Optional[t.Callable[[t.Any], t.Any]]


class Serializer:
    """Deserializes JSON data into wom.py model classes."""

    __slots__ = ()

    def _dt_from_iso(self, timestamp: str) -> datetime:
        return datetime.fromisoformat(timestamp.rstrip("Z"))

    def _dt_from_iso_maybe(self, timestamp: t.Optional[str]) -> t.Optional[datetime]:
        return self._dt_from_iso(timestamp) if timestamp else None

    def to_camel_case(self, attr: str) -> str:
        first, *rest = attr.split("_")
        return "".join((first.lower(), *map(str.title, rest)))

    def _set_attrs(
        self,
        model: t.Any,
        data: DictT,
        *attrs: str,
        transform: TransformT = None,
        camel_case: bool = False,
        maybe: bool = False,
    ) -> None:
        if transform and maybe:
            raise RuntimeError("Only one of 'maybe' and 'transform' may be used.")

        for attr in attrs:
            cased_attr = self.to_camel_case(attr) if camel_case else attr

            if transform:
                setattr(
                    model,
                    attr,
                    transform(data.get(cased_attr, None) if maybe else data[cased_attr]),
                )
            else:
                setattr(model, attr, data.get(cased_attr, None) if maybe else data[cased_attr])

    def _set_attrs_cased(
        self,
        model: t.Any,
        data: DictT,
        *attrs: str,
        transform: TransformT = None,
        maybe: bool = False,
    ) -> None:
        self._set_attrs(model, data, *attrs, transform=transform, camel_case=True, maybe=maybe)

    def to_api_key(self, data: DictT) -> models.ApiKey:
        model = models.ApiKey()
        self._set_attrs_cased(model, data, "key", "key_id")
        return model

    def to_api_key_verification(self, data: DictT) -> models.ApiKeyVerification:
        model = models.ApiKeyVerification()
        model.id = data.get("keyId")

        ratelimit = data.get("ratelimit")
        model.ratelimit = self.to_ratelimit_state(ratelimit) if ratelimit else ratelimit

        refill = data.get("refill")
        model.refill = self.to_refill(refill) if refill else refill

        model.code = models.ErrorCode.from_str_maybe(data.get("code", ""))
        self._set_attrs_cased(
            model, data, "valid", "owner_id", "meta", "remaining", "error", "expires", maybe=True
        )

        return model

    def to_ratelimit_state(self, data: DictT) -> models.RatelimitState:
        model = models.RatelimitState()
        self._set_attrs(model, data, "reset", "limit", "remaining")
        return model

    def to_api(self, data: DictT) -> models.Api:
        model = models.Api()
        self._set_attrs_cased(model, data, "id", "name", "workspace_id")
        return model

    def to_ratelimit(self, data: DictT) -> models.Ratelimit:
        return models.Ratelimit(
            limit=data["limit"],
            refill_rate=data["refillRate"],
            refill_interval=data["refillInterval"],
            type=models.RatelimitType.from_str(data["type"]),
        )

    def to_api_key_meta(self, data: DictT) -> models.ApiKeyMeta:
        model = models.ApiKeyMeta()

        ratelimit = data.get("ratelimit")
        model.ratelimit = self.to_ratelimit(ratelimit) if ratelimit else ratelimit

        refill = data.get("refill")
        model.refill = self.to_refill(refill) if refill else refill

        self._set_attrs_cased(
            model,
            data,
            "id",
            "meta",
            "start",
            "api_id",
            "expires",
            "remaining",
            "owner_id",
            "created_at",
            "workspace_id",
            maybe=True,
        )

        return model

    def to_api_key_list(self, data: DictT) -> models.ApiKeyList:
        model = models.ApiKeyList()
        model.cursor = data.get("cursor")
        model.total = data["total"]
        model.keys = [self.to_api_key_meta(key) for key in data["keys"]]
        return model

    def to_refill(self, data: DictT) -> models.Refill:
        interval = models.RefillInterval.from_str(data["interval"])
        amount = data["amount"]

        model = models.Refill(amount, interval)
        model.last_refilled_at = data.get("lastRefilledAt")

        return model



from __future__ import annotations

import typing as t

from unkey import errors
from unkey import models
from unkey import result
from unkey import routes
from unkey.undefined import UNDEFINED
from unkey.undefined import UndefinedNoneOr
from unkey.undefined import UndefinedOr
from unkey.undefined import all_undefined

from . import BaseService

__all__ = ("KeyService",)

T = t.TypeVar("T")
ResultT = result.Result[T, models.HttpResponse]


class KeyService(BaseService):
    """Handles api key related requests."""

    __slots__ = ()

    async def create_key(
        self,
        api_id: str,
        owner_id: str,
        prefix: str,
        *,
        name: UndefinedOr[str] = UNDEFINED,
        byte_length: UndefinedOr[int] = UNDEFINED,
        meta: UndefinedOr[t.Dict[str, t.Any]] = UNDEFINED,
        expires: UndefinedOr[int] = UNDEFINED,
        remaining: UndefinedOr[int] = UNDEFINED,
        ratelimit: UndefinedOr[models.Ratelimit] = UNDEFINED,
        refill: UndefinedOr[models.Refill] = UNDEFINED,
    ) -> ResultT[models.ApiKey]:
        """Creates a new api key.

        Args:
            name: The name to use for this key.

            api_id: The id of the api this key is for.

            owner_id: The owner id to use for this key. Represents the
                user who will use this key.

            prefix: The prefix to place at the beginning of the key.

        Keyword Args:
            byte_length: The optional desired length of they in bytes.
                Defaults to 16.

            meta: An optional dynamic mapping of information used to
                provide context around this keys user.

            expires: The optional number of milliseconds into the future
                when this key should expire.

            remaining: The optional max number of times this key can be
                used. Useful for creating long lived keys but with a
                limit on total uses.

            ratelimit: The optional `Ratelimit` to set on this key.

            refill: The optional `Refill` to set on this key.

        Returns:
            A result containing the newly created key or an error.
        """
        route = routes.CREATE_KEY.compile()
        payload = self._generate_map(
            meta=meta,
            name=name,
            apiId=api_id,
            prefix=prefix,
            ownerId=owner_id,
            remaining=remaining,
            byteLength=byte_length,
            expires=self._expires_in(milliseconds=expires or 0),
            ratelimit=UNDEFINED
            if not ratelimit
            else self._generate_map(
                limit=ratelimit.limit,
                type=ratelimit.type.value,
                refillRate=ratelimit.refill_rate,
                refillInterval=ratelimit.refill_interval,
            ),
            refill=UNDEFINED
            if not refill
            else self._generate_map(
                amount=refill.amount,
                interval=refill.interval.value,
            ),
        )

        data = await self._http.fetch(route, payload=payload)

        if isinstance(data, models.HttpResponse):
            return result.Err(data)

        return result.Ok(self._serializer.to_api_key(data))

    async def verify_key(self, key: str, api_id: str) -> ResultT[models.ApiKeyVerification]:
        """Verifies a key is valid and within ratelimit.

        Args:
            key: The key to verify.

            api_id: The id of the api to verify the key against.

        Returns:
            A result containing the api key verification or an error.
        """
        route = routes.VERIFY_KEY.compile()
        payload = self._generate_map(key=key, apiId=api_id)
        data = await self._http.fetch(route, payload=payload)

        if isinstance(data, models.HttpResponse):
            return result.Err(data)

        return result.Ok(self._serializer.to_api_key_verification(data))

    async def revoke_key(self, key_id: str) -> ResultT[models.HttpResponse]:
        """Revokes a keys validity.

        Args:
            key_id: The id of the key to revoke.

        Returns:
            A result containing the http response or an error.
        """
        route = routes.REVOKE_KEY.compile()
        payload = self._generate_map(keyId=key_id)
        data = await self._http.fetch(route, payload=payload)

        if isinstance(data, models.HttpResponse):
            return result.Err(data)

        return result.Ok(models.HttpResponse(200, "OK"))

    async def update_key(
        self,
        key_id: str,
        *,
        name: UndefinedNoneOr[str] = UNDEFINED,
        owner_id: UndefinedNoneOr[str] = UNDEFINED,
        meta: UndefinedNoneOr[t.Dict[str, t.Any]] = UNDEFINED,
        expires: UndefinedNoneOr[int] = UNDEFINED,
        remaining: UndefinedNoneOr[int] = UNDEFINED,
        ratelimit: UndefinedNoneOr[models.Ratelimit] = UNDEFINED,
        refill: UndefinedOr[models.Refill] = UNDEFINED,
    ) -> ResultT[models.HttpResponse]:
        """Updates an existing api key. To delete a key set its value
        to `None`.

        Args:
            key_id: The id of the key to update.

        Keyword Args:
            name: The new name to use for this key.

            owner_id: The new owner id to use for this key.

            meta: The new dynamic mapping of information used
                to provide context around this keys user.

            expires: The new number of milliseconds into the future
                when this key should expire.

            remaining: The new max number of times this key can be
                used.

            ratelimit: The new `Ratelimit` to set on this key.

            refill: The optional `Refill` to set on this key.

        Returns:
            A result containing the OK response or an error.
        """
        if all_undefined(name, owner_id, meta, expires, remaining, ratelimit, refill):
            raise errors.MissingRequiredArgument("At least one value is required to be updated.")

        route = routes.UPDATE_KEY.compile()
        payload = self._generate_map(
            name=name,
            meta=meta,
            keyId=key_id,
            ownerId=owner_id,
            remaining=remaining,
            expires=self._expires_in(milliseconds=expires or 0),
            ratelimit=UNDEFINED
            if not ratelimit
            else self._generate_map(
                limit=ratelimit.limit,
                type=ratelimit.type.value,
                refillRate=ratelimit.refill_rate,
                refillInterval=ratelimit.refill_interval,
            ),
            refill=UNDEFINED
            if not refill
            else self._generate_map(
                amount=refill.amount,
                interval=refill.interval.value,
            ),
        )

        data = await self._http.fetch(route, payload=payload)

        if isinstance(data, models.HttpResponse):
            return result.Err(data)

        return result.Ok(models.HttpResponse(200, "OK"))

    async def get_key(self, key_id: str) -> ResultT[models.ApiKeyMeta]:
        """Retrieves details for the given key.

        Args:
            key_id: The id of the key.

        Returns:
            A result containing the api key metadata or an error.
        """
        params = self._generate_map(keyId=key_id)
        route = routes.GET_KEY.compile().with_params(params)
        data = await self._http.fetch(route)

        if isinstance(data, models.HttpResponse):
            return result.Err(data)

        return result.Ok(self._serializer.to_api_key_meta(data))

    async def update_remaining(
        self, key_id: str, value: t.Optional[int], op: models.UpdateOp
    ) -> ResultT[t.Optional[int]]:
        """Updates a keys remaining limit.

        Args:
            key_id: The id of the key.

            value: The value to perform the operation on.

            op: The update operation to perform.

        Returns:
            A result containing the new remaining limit of the key or an error.
        """
        payload = self._generate_map(keyId=key_id, value=value, op=op.value)
        route = routes.UPDATE_REMAINING.compile()
        data = await self._http.fetch(route, payload=payload)

        if isinstance(data, models.HttpResponse):
            return result.Err(data)

        return result.Ok(data["remaining"])



from __future__ import annotations

import typing as t

import attrs

from .base import BaseEnum
from .base import BaseModel
from .http import ErrorCode

__all__ = (
    "ApiKey",
    "ApiKeyMeta",
    "ApiKeyVerification",
    "Ratelimit",
    "RatelimitState",
    "RatelimitType",
    "Refill",
    "RefillInterval",
    "UpdateOp",
)


class RatelimitType(BaseEnum):
    Fast = "fast"
    Consistent = "consistent"


class RefillInterval(BaseEnum):
    Daily = "daily"
    Monthly = "monthly"


class UpdateOp(BaseEnum):
    Increment = "increment"
    Decrement = "decrement"
    Set = "set"


@attrs.define(weakref_slot=False)
class Ratelimit(BaseModel):
    """Data representing a particular ratelimit."""

    type: RatelimitType
    """The type of ratelimiting to implement."""

    limit: int
    """The total amount of burstable requests."""

    refill_rate: int
    """How many tokens to refill during each refill interval."""

    refill_interval: int
    """The speed at which tokens are refilled."""


@attrs.define(init=False, weakref_slot=False)
class ApiKey(BaseModel):
    """Minimal representation of an api key."""

    key_id: str
    """The id of this key stored at unkey."""

    key: str
    """The api key itself."""


@attrs.define(init=False, weakref_slot=False)
class ApiKeyMeta(BaseModel):
    """Metadata about an api key."""

    id: str
    """The id of this key."""

    api_id: str
    """The id of the api this key belongs to."""

    workspace_id: str
    """The id of the workspace this key belongs to."""

    start: str
    """The prefix and beginning 3 letters of the key."""

    created_at: int
    """The unix epoch representing when this key was created in
    milliseconds."""

    owner_id: t.Optional[str]
    """The owner of this api key if one was specified."""

    expires: t.Optional[int]
    """The optional unix epoch representing when this key expires in
    milliseconds."""

    ratelimit: t.Optional[Ratelimit]
    """The optional ratelimit associated with this key."""

    meta: t.Optional[t.Dict[str, t.Any]]
    """The dynamic mapping of data used during key creation, if
    the key was found.
    """

    remaining: t.Optional[int]
    """The remaining verifications before this key is invalidated.
    If `None`, this field was not used in the keys creation and can
    be ignored.
    """

    refill: t.Optional[Refill]
    """The keys refill state, if any."""


@attrs.define(init=False, weakref_slot=False)
class ApiKeyVerification(BaseModel):
    """Data about whether this api key and its validity."""

    id: t.Optional[str]
    """The id of this key."""

    valid: bool
    """Whether or not this key is valid and passes ratelimit."""

    owner_id: t.Optional[str]
    """The id of the owner for this key, if the key was found."""

    meta: t.Optional[t.Dict[str, t.Any]]
    """Dynamic mapping of data used during key creation, if the
    key was found.
    """

    remaining: t.Optional[int]
    """The remaining verifications before this key is invalidated.
    If `None`, this field was not used in the keys creation and can
    be ignored.
    """

    expires: t.Optional[int]
    """The unix epoch in milliseconds indicating when this key expires,
    if it does."""

    ratelimit: t.Optional[RatelimitState]
    """The state of the ratelimit set on this key, if any."""

    refill: t.Optional[Refill]
    """The keys refill state, if any."""

    code: t.Optional[ErrorCode]
    """The optional error code returned by the unkey api."""

    error: t.Optional[str]
    """The error message if the key was invalid."""

    @staticmethod
    def __internal_serialize(_: type, __: t.Any, value: t.Any) -> t.Any:
        if isinstance(value, ErrorCode):
            # Automatically convert ErrorCodes to strings for to_dict()
            return str(value)

        return value

    def to_dict(self) -> t.Dict[str, t.Any]:
        """Converts this class into a dictionary.

        Returns:
            The requested dictionary.
        """
        return attrs.asdict(self, value_serializer=self.__internal_serialize)


@attrs.define(init=False, weakref_slot=False)
class RatelimitState(BaseModel):
    """The state of the ratelimit for a given key."""

    limit: int
    """The number of burstable requests allowed."""

    remaining: int
    """The remaining requests in this burst window."""

    reset: int
    """The unix timestamp in milliseconds until the next window."""


@attrs.define(weakref_slot=False)
class Refill(BaseModel):
    """Data regarding how a key's verifications should be refilled."""

    amount: int
    """The number of verifications to refill."""

    interval: RefillInterval
    """The interval at which to refill the verifications."""

    last_refilled_at: t.Optional[int] = None
    """The UNIX timestamp in milliseconds indicating when the key was
    las refilled, if it has been.
    """



from __future__ import annotations

import typing as t

import attrs

from .base import BaseEnum
from .base import BaseModel

__all__ = ("ErrorCode", "HttpResponse")


class ErrorCode(BaseEnum):
    NotFound = "NOT_FOUND"
    BadRequest = "BAD_REQUEST"
    Unauthorized = "UNAUTHORIZED"
    InternalServerError = "INTERNAL_SERVER_ERROR"
    Ratelimited = "RATELIMITED"
    Forbidden = "FORBIDDEN"
    KeyUsageExceeded = "KEY_USAGE_EXCEEDED"
    InvalidKeyType = "INVALID_KEY_TYPE"
    NotUnique = "NOT_UNIQUE"
    Unknown = "UNKNOWN"
    Conflict = "CONFLICT"


@attrs.define(weakref_slot=False)
class HttpResponse(BaseModel):
    """Directly represents a response from the api.

    Could indicate either success or failure.
    """

    status: int
    """The HTTP status code."""

    message: str
    """The error or success message."""

    code: t.Optional[ErrorCode] = None
    """The optional error code returned by the unkey api."""



from __future__ import annotations

import typing as t

from unkey import models
from unkey import result
from unkey import routes
from unkey.undefined import UNDEFINED
from unkey.undefined import UndefinedOr

from . import BaseService

__all__ = ("ApiService",)

T = t.TypeVar("T")
ResultT = result.Result[T, models.HttpResponse]


class ApiService(BaseService):
    """Handles api related requests."""

    __slots__ = ()

    async def get_api(self, api_id: str) -> ResultT[models.Api]:
        """Gets information about an api.

        Args:
            api_id: The id of the api.

        Returns:
            A result containing the requested information or an error.
        """
        params = self._generate_map(apiId=api_id)
        route = routes.GET_API.compile().with_params(params)
        data = await self._http.fetch(route)

        if isinstance(data, models.HttpResponse):
            return result.Err(data)

        return result.Ok(self._serializer.to_api(data))

    async def list_keys(
        self,
        api_id: str,
        *,
        owner_id: UndefinedOr[str] = UNDEFINED,
        limit: UndefinedOr[int] = UNDEFINED,
        cursor: UndefinedOr[str] = UNDEFINED,
    ) -> ResultT[models.ApiKeyList]:
        """Gets a paginated list of keys for the given api.

        Args:
            api_id: The id of the api.

        Keyword Args:
            owner_id: The optional owner id to list the keys for.

            limit: The optional max number of keys to include in this page.

            cursor: Optional key used to determine pagination offset.

        Returns:
            A result containing api key list or an error.
        """
        params = self._generate_map(apiId=api_id, ownerId=owner_id, limit=limit, cursor=cursor)
        route = routes.GET_KEYS.compile().with_params(params)
        data = await self._http.fetch(route)

        if isinstance(data, models.HttpResponse):
            return result.Err(data)

        return result.Ok(self._serializer.to_api_key_list(data))



from __future__ import annotations

import abc
import typing as t

from unkey import errors

__all__ = ("Err", "Ok", "Result")

T = t.TypeVar("T")
E = t.TypeVar("E")


class Result(t.Generic[T, E], abc.ABC):
    """Represents a potential `Ok` or `Err` result.

    Note:
        This class can not be instantiated, only its variants can.
    """

    __slots__ = ("_error", "_value")

    def __repr__(self) -> str:
        inner = self._value if self.is_ok else self._error  # type: ignore [attr-defined]
        return f"{self.__class__.__name__}({inner})"

    @property
    @abc.abstractmethod
    def is_ok(self) -> bool:
        """`True` if this result is the `Ok` variant."""

    @property
    @abc.abstractmethod
    def is_err(self) -> bool:
        """`True` if this result is the `Err` variant."""

    @abc.abstractmethod
    def unwrap(self) -> T:
        """Unwraps the result to produce the value.

        Returns:
            The unwrapped value.

        Raises:
            UnwrapError: If the result was an `Err` and not `Ok`.
        """

    @abc.abstractmethod
    def unwrap_err(self) -> E:
        """Unwraps the result to produce the error.

        Returns:
            The unwrapped error.

        Raises:
            UnwrapError: If the result was `Ok` and not an `Err`.
        """


@t.final
class Ok(Result[T, E]):
    __slots__ = ()

    def __init__(self, value: T) -> None:
        self._value = value

    @property
    def is_ok(self) -> bool:
        return True

    @property
    def is_err(self) -> bool:
        return False

    def unwrap(self) -> T:
        return self._value

    def unwrap_err(self) -> E:
        actual = self._value.__class__.__name__
        raise errors.UnwrapError(f"Called unwrap error on a non error value of type {actual!r}")


@t.final
class Err(Result[T, E]):
    __slots__ = ()

    def __init__(self, error: E) -> None:
        self._error = error

    @property
    def is_ok(self) -> bool:
        return False

    @property
    def is_err(self) -> bool:
        return True

    def unwrap(self) -> T:
        raise errors.UnwrapError(f"Called unwrap on an error value - {self._error}")

    def unwrap_err(self) -> E:
        return self._error



from __future__ import annotations

import typing as t
from enum import Enum

import attrs

__all__ = ("BaseEnum", "BaseModel")

T = t.TypeVar("T", bound="BaseEnum")


@attrs.define(weakref_slot=False)
class BaseModel:
    """The base model all library models inherit from."""

    def to_dict(self) -> t.Dict[str, t.Any]:
        """Converts this class into a dictionary.

        Returns:
            The requested dictionary.
        """
        return attrs.asdict(self)


class BaseEnum(Enum):
    """The base enum all library enums inherit from."""

    __slots__ = ()

    value: str  # pyright: ignore

    def __str__(self) -> str:
        return self.value

    @classmethod
    def from_str(cls: t.Type[T], value: str) -> T:
        """Generate this enum from the given value.

        Args:
            value: The value to generate from.

        Returns:
            The generated enum.
        """
        try:
            return cls(value)
        except ValueError as e:
            raise ValueError(
                f"{e} variant. Please report this issue on github at "
                "https://github.com/Jonxslays/unkey.py/issues/new"
            ) from None

    @classmethod
    def from_str_maybe(cls: t.Type[T], value: t.Optional[str]) -> t.Optional[T]:
        """Attempt to generate this enum from the given value.

        Args:
            value: The value to generate from.

        Returns:
            The generated enum or `None` if the value was not a valid
                enum variant.
        """
        try:
            return cls(value)
        except ValueError:
            return None



from __future__ import annotations

import typing as t

__all__ = ("UndefinedNoneOr", "UndefinedOr", "UNDEFINED")


class Undefined:
    """Represents an undefined value - without being None."""

    __slots__ = ()

    def __bool__(self) -> t.Literal[False]:
        return False

    def __copy__(self) -> Undefined:
        return self

    def __deepcopy__(self, memo: t.MutableMapping[int, t.Any]) -> Undefined:
        memo[id(self)] = self
        return self

    def __getstate__(self) -> t.Any:
        return False

    def __repr__(self) -> str:
        return "UNDEFINED"

    def __reduce__(self) -> str:
        return "UNDEFINED"

    def __str__(self) -> str:
        return "UNDEFINED"


def __singleton_new(cls: t.Any) -> t.NoReturn:
    raise TypeError("Cannot instantiate singleton class UNDEFINED.")


UNDEFINED = Undefined()
"""A value that does not exist."""

Undefined.__new__ = __singleton_new  # type: ignore

T = t.TypeVar("T", covariant=True)

UndefinedOr = t.Union[T, Undefined]
"""A value that is undefined or T"""

UndefinedNoneOr = UndefinedOr[t.Optional[T]]
"""A value that is undefined, none, or T"""


def all_undefined(*values: t.Any) -> bool:
    """Whether or not all values are undefined.

    Arguments:
        *values: The values to check.

    Returns:
        `True` if all values were undefined.
    """
    return all(v is UNDEFINED for v in values)


def any_undefined(*values: t.Any) -> bool:
    """Whether or not any values are undefined.

    Arguments:
        *values: The values to check.

    Returns:
        `True` if any values were undefined.
    """
    return any(v is UNDEFINED for v in values)



from .base import *

from .apis import *
from .http import *
from .keys import *

__all__ = (
    "Api",
    "ApiKey",
    "ApiKeyList",
    "ApiKeyMeta",
    "ApiKeyVerification",
    "BaseEnum",
    "BaseModel",
    "ErrorCode",
    "HttpResponse",
    "Ratelimit",
    "RatelimitState",
    "RatelimitType",
    "Refill",
    "RefillInterval",
    "UpdateOp",
)



from __future__ import annotations

__all__ = ("BaseError", "MissingRequiredArgument", "UnwrapError")


class BaseError(Exception):
    """The base error all unkey errors inherit from."""

    __slots__ = ()


class UnwrapError(BaseError):
    """Raised when calling unwrap or unwrap_err incorrectly."""

    __slots__ = ()

    def __init__(self, message: str) -> None:
        super().__init__(f"Unwrap failed: {message}")


class MissingRequiredArgument(BaseError):
    """Raised when a required argument is missing."""

    __slots__ = ()

    def __init__(self, message: str) -> None:
        super().__init__(f"Missing required argument: {message}")



from __future__ import annotations

import typing as t

import aiohttp

from unkey import constants
from unkey import models
from unkey import routes

__all__ = ("HttpService",)

T = t.TypeVar("T")


class HttpService:
    """The HTTP service used to make requests to the unkey API.

    Args:
        api_key: The api key to use.

        api_version: The optional version of the api to use.

        api_base_url: The optional api base url to use.
    """

    __slots__ = (
        "_api_version",
        "_base_url",
        "_headers",
        "_ok_responses",
        "_method_mapping",
        "_session",
    )

    def __init__(
        self,
        api_key: t.Optional[str],
        api_version: t.Optional[int],
        api_base_url: t.Optional[str],
    ) -> None:
        if api_key == "":
            raise ValueError("Api key must not be empty.")

        self._headers = {
            "Unkey-SDK": constants.USER_AGENT,
            "User-Agent": constants.USER_AGENT,
            "x-user-agent": constants.USER_AGENT,
        }

        if api_key:
            self._headers["Authorization"] = f"Bearer {api_key}"

        self._ok_responses = {200, 202}
        self._api_version = f"/v{api_version or 1}"
        self._base_url = api_base_url or constants.API_BASE_URL

    async def _try_get_json(self, response: aiohttp.ClientResponse) -> t.Any:
        try:
            return await response.json()
        except Exception:
            if response.status not in self._ok_responses:
                return models.HttpResponse(response.status, await response.text())

            return await response.text()

    async def _request(
        self, req: t.Callable[..., t.Awaitable[t.Any]], url: str, **kwargs: t.Any
    ) -> t.Any:
        response = await req(url, **kwargs)
        data = await self._try_get_json(response)

        if isinstance(data, models.HttpResponse):
            return data

        if response.status not in self._ok_responses:
            error: t.Union[t.Any, t.Dict[str, t.Any]] = data.get("error")
            is_dict = isinstance(error, dict)

            message = error.get("message") if is_dict else error
            code = models.ErrorCode.from_str_maybe(error.get("code") if is_dict else "UNKNOWN")
            message = message or "An unexpected error occurred while making the request."

            return models.HttpResponse(response.status, str(message), code=code)

        return data

    def _get_request_func(self, method: str) -> t.Callable[..., t.Awaitable[t.Any]]:
        if not hasattr(self, "_method_mapping"):
            raise RuntimeError("HttpService.start was never called, aborting...")

        return self._method_mapping[method]  # type: ignore

    async def _init_session(self) -> None:
        self._session = aiohttp.ClientSession()
        self._method_mapping = {
            constants.GET: self._session.get,
            constants.PUT: self._session.put,
            constants.POST: self._session.post,
            constants.PATCH: self._session.patch,
            constants.DELETE: self._session.delete,
        }

    def set_api_key(self, api_key: str) -> None:
        """Sets the api key used by the http service.

        Args:
            api_key: The new api key to use.
        """
        self._headers["x-api-key"] = api_key

    def set_base_url(self, base_url: str) -> None:
        """Sets the api base url used by the http service.

        Args:
            base_url: The new base url to use.
        """
        self._base_url = base_url

    async def start(self) -> None:
        """Starts the client session to be used by the http service."""
        if not hasattr(self, "_session") or self._session.closed:
            await self._init_session()

    async def close(self) -> None:
        """Closes the existing client session, if it's still open."""
        if hasattr(self, "_session") and not self._session.closed:
            await self._session.close()

    async def fetch(
        self,
        route: routes.CompiledRoute,
        *,
        payload: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> dict[str, t.Any] | models.HttpResponse:
        """Fetches the given route.

        Args:
            route: The route to make the request to.

            payload: The optional payload to send in the request body.

        Returns:
            The requested json data or the error response.
        """
        return await self._request(  # type: ignore[no-any-return]
            self._get_request_func(route.method),
            self._base_url + self._api_version + route.uri,
            headers=self._headers,
            params=route.params,
            json=payload or None,
        )



from __future__ import annotations

import typing as t

from unkey import serializer
from unkey import services

__all__ = ("Client",)

ServiceT = t.TypeVar("ServiceT")


class Client:
    """An asynchronous client used for interacting with the API.

    Args:
        api_key: The root api key to use for requests.

    Keyword Args:
        api_version: The api version to access. Defaults to 1.

        api_base_url: The base url to use for the api (no trailing /).
            Defaults to `https://api.unkey.dev`.
    """

    __slots__ = (
        "_apis",
        "_http",
        "_keys",
        "_serializer",
    )

    def __init__(
        self,
        api_key: t.Optional[str] = None,
        *,
        api_version: t.Optional[int] = None,
        api_base_url: t.Optional[str] = None,
    ) -> None:
        self._serializer = serializer.Serializer()
        self._http = services.HttpService(api_key, api_version, api_base_url)
        self.__init_core_services()

    def __init_core_services(self) -> None:
        self._apis = self.__init_service(services.ApiService)
        self._keys = self.__init_service(services.KeyService)

    def __init_service(self, service: t.Type[ServiceT]) -> ServiceT:
        if not issubclass(service, services.BaseService):
            raise TypeError(f"{service.__name__!r} can not be initialized as a service.")

        return service(self._http, self._serializer)  # type: ignore[return-value]

    @property
    def keys(self) -> services.KeyService:
        """The key service used to make key related requests."""
        return self._keys

    @property
    def apis(self) -> services.ApiService:
        """The api service used to make api related requests."""
        return self._apis

    def set_api_key(self, api_key: str) -> None:
        """Sets the api key used by the http service.

        Args:
            api_key: The new root api key to use for requests.
        """
        self._http.set_api_key(api_key)

    def set_api_base_url(self, base_url: str) -> None:
        """Sets the api base url used by the http service.

        Args:
            base_url: The new api base url to use for requests.
        """
        self._http.set_base_url(base_url)

    async def start(self) -> None:
        """Starts the client session to be used for http requests."""
        await self._http.start()

    async def close(self) -> None:
        """Closes the existing client session, if it's still open."""
        await self._http.close()



from __future__ import annotations

import functools
import inspect
from typing import Any
from typing import Callable
from typing import Coroutine
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import TypeVar
from typing import Union

from unkey import client
from unkey import models

__all__ = ("protected",)

T = TypeVar("T")
CallableT = Callable[..., T]
CallbackT = CallableT[Any]
DecoratorT = CallableT[CallbackT]
VerificationResponseT = Union[T, Any]
ExtractorT = Callable[[Tuple[Any], Dict[str, Any]], Optional[str]]
"""The type of a callback used to extract the api key from the decorated
functions `*args` and `**kwargs`.
"""

InvalidKeyHandlerT = Optional[Callable[[Dict[str, Any], Optional[models.ApiKeyVerification]], Any]]
"""The type of a callback used to handle cases where the key was invalid."""

ExcHandlerT = Optional[Callable[[Exception], Any]]
"""The type of a callback used to handle exceptions during verification."""


def protected(
    api_id: str,
    key_extractor: ExtractorT,
    on_invalid_key: InvalidKeyHandlerT = None,
    on_exc: ExcHandlerT = None,
) -> DecoratorT:
    """A framework agnostic second order decorator that is used to protect
    api routes with Unkey key verification.

    !!! info

        An endpoint decorated with this decorator guarantees that once the
        decorated function body is reached that the api key was verified and
        is valid.

        A keyword argument `unkey_verification` will be injected into the
        function on a successful verification so you should make sure you
        accept that argument in the decorated function. From there feel
        free to inspect the returned `ApiKeyVerification` for any properties
        of interest.

    Args:
        api_id: The ID of the api to verify keys against.

        key_extractor: The callback function used to extract the api key
            from the `*args` and `**kwargs` of the decorated function. It
            should return the key stripped of the "Bearer " prefix or `None`.

        on_invalid_key: The callback function used to transform the output
            from an invalid key. This can mean the key could not be parsed, the
            key was not found, or the key was actually invalid. This function
            should accept a dictionary and an optional `ApiKeyVerification`.
            The verification will be provided in cases where we were able to
            successfully retrieve it from Unkey.

        on_exc: The callback function used to handle exceptions that get thrown
            at any point during verification.

    Raises:
        exc: If an exception is raised and no `on_exc` callback was supplied.

    Returns:
        A function that replaces the decorated function, and performs Unkey
            key verification prior to calling the decorated function.
            If the key could not be parsed or was invalid, a dictionary
            containing "code" and "message" keys will be returned, unless
            altered by `on_invalid_key` if it was passed. If verification
            succeeds the original functions return value is returned.
    """
    _client = client.Client()

    def _on_invalid_key(
        data: Dict[str, Any], verification: Optional[models.ApiKeyVerification] = None
    ) -> Any:
        if on_invalid_key:
            return on_invalid_key(data, verification)

        return data

    def _on_exc(exc: Exception) -> Any:
        if on_exc:
            return on_exc(exc)

        raise exc

    def wrapper(
        func: CallableT[T],
    ) -> CallableT[Coroutine[Any, Any, VerificationResponseT[T]]]:
        @functools.wraps(func)
        async def inner(*args: Any, **kwargs: Any) -> VerificationResponseT[T]:
            try:
                if not (key := key_extractor(*args, **kwargs)):
                    message = "Failed to extract API key"
                    return _on_invalid_key({"code": None, "message": message})

                await _client.start()
                result = await _client.keys.verify_key(key, api_id)
                await _client.close()

                if result.is_err:
                    err = result.unwrap_err()
                    code = (err.code or models.ErrorCode.Unknown).value
                    return _on_invalid_key({"code": code, "message": err.message})

                verification = result.unwrap()
                kwargs["unkey_verification"] = verification

                if not verification.valid:
                    code = (verification.code or models.ErrorCode.Unknown).value
                    return _on_invalid_key(
                        {"code": code, "message": verification.error}, verification
                    )

                if inspect.iscoroutinefunction(func):
                    value = await func(*args, **kwargs)
                else:
                    value = func(*args, **kwargs)

            except Exception as exc:
                return _on_exc(exc)

            return value

        return inner

    return wrapper
